/* zselx - interactively select files and execute command.
 * Public domain, by Russell Marks.
 */

#define VERSION	"1.1"

#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <ctype.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/param.h>
#include <time.h>
#include <curses.h>


#define MAX_FILES	1024

#define MAX_FILE_LINES	(LINES-3)

#ifdef A_NORMAL
#define SYSV_CURSES
#else
/* define keys so they don't get in the way */
#define NOT_A_KEY	0x7fffffff
#define KEY_UP		NOT_A_KEY
#define KEY_DOWN	NOT_A_KEY
#define KEY_LEFT	NOT_A_KEY
#define KEY_RIGHT	NOT_A_KEY
#define KEY_NPAGE	NOT_A_KEY
#define KEY_PPAGE	NOT_A_KEY
#define KEY_HOME	NOT_A_KEY
#define KEY_LL		NOT_A_KEY
#endif


struct files_tag {
  int tagged;
  char name[256];
  int extofs;		/* offset of 'extension' in name[] */
  time_t date;
  int size;
  int mode;
  } files[MAX_FILES];

int numfiles;

/* values for cmdtype.
 * I can't use plain 'move' because curses defines 'move', so I
 * give that a prefix. I've used prefixes for other enum bits for
 * consistency, even if it is a bit icky. (I mean, I might as well
 * be using #define's with it like this. :-/)
 */
enum cmdtypes
  {
  cmd_none,
  cmd_copy,cmd_move,cmd_delete,
  cmd_view,cmd_edit,
  cmd_other
  };

enum sorttypes { sort_name,sort_ext,sort_size,sort_date };

/* this has to be global so we can use it in mfl_sortcmp() */
enum sorttypes sorttype=sort_name;

int opt_all=0,opt_one=0;


/* XXX hardly optimal - should use {insert,delete}ln() when top has
 * only moved one rather than redrawing the lot.
 * Ok, this won't matter much on a console or xterm, but... :-)
 */
void drawfilelist(int top,int pos,int oldtop,int oldpos)
{
int startline=3;
int f;
char buf[256];
struct files_tag *fp;

/* don't bother if nothing's changed */
if(top!=oldtop || pos!=oldpos)
  {
  for(f=0;f<MAX_FILE_LINES;f++)
    {
    move(startline+f,0);
    clrtoeol();
    if(top+f>=numfiles) continue;
    
    move(startline+f,0);
    fp=files+top+f;
    if(fp->tagged) addstr("* "); else addstr("  ");
    strcpy(buf,fp->name);
    if(COLS>=40)
      {
      struct tm *tp=localtime(&(fp->date));
      
      printw("%6dk  %4d/%02d/%02d %2d:%02d  ",(fp->size+1023)/1024,
        tp->tm_year+1900,tp->tm_mon+1,tp->tm_mday,
        tp->tm_hour,tp->tm_min);
      
      if(COLS-29>0 && COLS-29<256) buf[COLS-29]=0;
      }
    
    if(opt_all)
      {
      if(S_ISLNK(fp->mode))	strcat(buf,"@"); else
      if(S_ISDIR(fp->mode))	strcat(buf,"/"); else
      if(S_ISFIFO(fp->mode))	strcat(buf,"|"); else
      if(S_ISSOCK(fp->mode))	strcat(buf,"="); else
      /* yes, nothing for devices - that's what 'ls -F' does too */
      if(fp->mode & 0111)	strcat(buf,"*");
      }
    
    addstr(buf);
    }
  }

/* we always move the cursor to the right place, though... */
move(startline+pos-top,0);
}


int mfl_sortcmp(void *p1,void *p2)
{
struct files_tag *f1=(struct files_tag *)p1;
struct files_tag *f2=(struct files_tag *)p2;
int ret=0;

switch(sorttype)
  {
  case sort_name:
    ret=strcmp(f1->name,f2->name);
    break;
  
  case sort_ext:
    ret=strcmp(f1->name+f1->extofs,f2->name+f2->extofs);
    break;
  
  case sort_size:
    if(f1->size<f2->size)
      ret=-1;
    else
      if(f1->size>f2->size)
        ret=1;
    break;
  
  case sort_date:
    if(f1->date<f2->date)
      ret=-1;
    else
      if(f1->date>f2->date)
        ret=1;
    break;
  }

/* for all equal matches on primary key, use name as secondary */
if(ret==0)
  ret=strcmp(f1->name,f2->name);

return(ret);
}


void resortlist(enum sorttypes sorttype)
{
qsort(&(files[0]),numfiles,sizeof(struct files_tag),(void *)mfl_sortcmp);
}


/* construct/sort file list */
void makefilelist(enum sorttypes sorttype)
{
DIR *dirfile;
struct dirent *anentry;
struct stat buf;
char cdir[MAXPATHLEN+1],*ptr;
int entnum;

getcwd(cdir,MAXPATHLEN);
dirfile=opendir(".");
entnum=0;
while((anentry=readdir(dirfile))!=NULL)
  {
  if(entnum==MAX_FILES) break;
  if(anentry->d_name[0]=='.')
    continue;	/* skip 'dot' files */
  /* we use lstat to get the right results when using '-a' */
  if((lstat(anentry->d_name,&buf))!=-1 &&
     (opt_all || (!opt_all && S_ISREG(buf.st_mode))))
    {
    strcpy(files[entnum].name,anentry->d_name);
    if((ptr=strrchr(files[entnum].name,'.'))==NULL)
      files[entnum].extofs=strlen(files[entnum].name);
    else
      files[entnum].extofs=ptr-files[entnum].name;
    files[entnum].date=buf.st_mtime;
    files[entnum].size=buf.st_size;
    files[entnum].mode=buf.st_mode;
    files[entnum].tagged=0;
    entnum++;
    }
  }
closedir(dirfile);
numfiles=entnum;
resortlist(sorttype);
}


/* show normal (usage) header:
 * this should be small enough to display on even really bogus terminals
 * like zcn/nc100 or cpm/+3, and covers most major opts.
 */
void showhdr()
{
move(0,0); clrtoeol();
move(1,0); clrtoeol();
mvaddstr(0,0,"zselx - tag/clr by +/-, A to clr all, Q quits");
mvaddstr(1,0,"cmds: Copy Move Del Other order: Name Ext Size Time");
mvaddstr(2,0,"---");
}


void usage_help()
{
printf("Zselx v%s - public domain; ",VERSION);
printf("by Russell Marks for improbabledesigns.\n\n");
printf("usage: zselx [-aho] [dir_path]\n\n");
printf("	-a	show all files rather than just `regular' ones.\n");
printf("	-h	give this usage help.\n");
printf("	-o	repeat; restart after each command.\n");
}


void parse_options(int argc,char **argv)
{
int doneopts=0;

opterr=0;

do
  switch(getopt(argc,argv,"aho"))
    {
    case 'a':	/* show all files */
      opt_all=1;
      break;
    case 'h':
      usage_help();
      exit(1);
    case 'o':	/* 'oneshot' - one command, then quit */
      opt_one=1;
      break;
    case '?':
      fprintf(stderr,"zselx: option '%c' not recognised.\n",optopt);
      exit(1);
    case -1:
      doneopts=1;
    }
while(!doneopts);

/* if more than one arg left, give usage */
if(optind<argc-1)
  {
  usage_help();
  exit(1);
  }

/* cd to dir if given */
if(optind==argc-1) chdir(argv[optind]);
}


void curses_init()
{
initscr();
cbreak(); noecho();

#ifdef SYSV_CURSES
keypad(stdscr,TRUE);
#endif
}


/* call this to re-init after calling curses_uninit() to temporarily
 * 'escape' from curses.
 */
void curses_reinit()
{
cbreak(); noecho();
refresh();
}


void curses_uninit()
{
clear(); refresh();
echo(); nocbreak();
endwin();
}


void do_select(enum cmdtypes *cmdtype_ptr,char *cmdtext)
{
int quit=0;
int autosel;
int top,pos,oldtop,oldpos;
char dest[1024],runcmd[1024];
enum cmdtypes cmdtype=cmd_none;
int f,c,tot;

clear();
showhdr();

top=pos=0;
oldtop=oldpos=-1;

while(!quit)
  {
  /* draw file list */
  drawfilelist(top,pos,oldtop,oldpos);
  
  oldtop=top; oldpos=pos;
  
  switch(getch())
    {
    /* quit */
    case 27: case 'x': case 'q':
      quit=1; break;
    case '1': case 'c':
      cmdtype=cmd_copy; break;
    case '2': case 'd':
      cmdtype=cmd_delete; break;
    case '3': case 'm':
      cmdtype=cmd_move; break;
    case 'o': case '6':
      cmdtype=cmd_other; break;
    case 'v':
      cmdtype=cmd_view; break;
    case 'f':
      cmdtype=cmd_edit; break;
    
    /* sort type */
    case '7': case 'n':
      if(sorttype!=sort_name) resortlist(sorttype=sort_name);
      oldtop=oldpos=-1;
      break;
    case '8': case 'e':
      if(sorttype!=sort_ext) resortlist(sorttype=sort_ext);
      oldtop=oldpos=-1;
      break;
    case '9': case 's':
      if(sorttype!=sort_size) resortlist(sorttype=sort_size);
      oldtop=oldpos=-1;
      break;
    case 't':	/* can't do '0' too, as used for vi movement cmd :-( */
      if(sorttype!=sort_date) resortlist(sorttype=sort_date);
      oldtop=oldpos=-1;
      break;
    
    /* movement */
    case 'h': case 'k': case 'p'-0x60: case KEY_UP: case KEY_LEFT:
      pos--; break;
    case 'l': case 'j': case 'n'-0x60: case KEY_DOWN: case KEY_RIGHT:
      pos++; break;
    case '0': case '<': case KEY_HOME:
      pos=0; break;
    /* not sure if KEY_END is in every sysv curses,
     * but ncurses needs it for the Linux console, at least.
     * (though this is presumably a consequence of the terminfo entry...?)
     */
#ifdef KEY_END
    case KEY_END:
#endif
    case '$': case 'G': case '>': case KEY_LL:
      pos=numfiles-1; break;
    case 'f'-0x60: case 'v'-0x60: case KEY_NPAGE:
      pos+=MAX_FILE_LINES-2; break;
    case 'b'-0x60: case 'u'-0x60: case KEY_PPAGE:
      pos-=MAX_FILE_LINES-2; break;
    
    /* tagging */
    case '4': case 'a':
      for(f=0;f<numfiles;f++)
        files[f].tagged=0;
      oldtop=oldpos=-1;
      break;
    case '5':
      /* XXX what does 'mark blank' do!? */
      break;
    
    /* need to force redisplay for these as it could be last file */
    case '+': case '=':
      files[pos++].tagged=1; oldpos=-1;
      break;
    case '-':
      files[pos++].tagged=0; oldpos=-1;
      break;
    case ' ':
      files[pos].tagged=!files[pos].tagged; oldpos=-1;
      pos++;
      break;
    case '~':
      for(f=0;f<numfiles;f++)
        if(strlen(files[f].name)>1 &&
        	files[f].name[strlen(files[f].name)-1]=='~')
          files[f].tagged=1,oldpos=-1;
      break;
    }
  
  if(pos<0) pos=0;
  if(pos>=numfiles) pos=numfiles-1;
  
  if(pos<top) top=pos;
  if(pos>top+MAX_FILE_LINES-1) top=pos-MAX_FILE_LINES+1;
  if(top<0) top=0;
  
  if(cmdtype!=cmd_none)
    {
    /* if they chose a command, check it's ok and/or get dest, etc. */
    autosel=0;
    
    /* if no files tagged, tag current */
    tot=0;
    for(f=0;f<numfiles;f++)
      if(files[f].tagged) tot++;
    if(tot==0) files[pos].tagged=autosel=1;
    
    /* clear header area bit, as all below use it */
    move(0,0); clrtoeol();
    move(1,0); clrtoeol();
    
    /* zero out dest in case it isn't defined below */
    dest[0]=0;
      
    switch(cmdtype)
      {
      case cmd_copy:
        strcpy(runcmd,"cp -pv");
        /* FALLS THROUGH */
      
      case cmd_move:
        if(cmdtype==cmd_move) strcpy(runcmd,"mv -v");
        
        /* get dest */
        mvaddstr(0,0,"Enter dir to ");
        addstr((cmdtype==cmd_copy)?"copy":"move");
        addstr(" files to, or none to abort:");
        echo(); mvgetstr(1,0,dest); noecho();
        if(dest[0]==0) cmdtype=cmd_none;
        break;
      
      case cmd_delete:
        strcpy(runcmd,"rm -fv");
        
        /* confirm */
        mvaddstr(0,0,"Delete marked files (y/n)? ");
        while((c=getch())!='y' && c!='n') ;
        if(c=='n') cmdtype=cmd_none;
        break;
      
      case cmd_view:
        if(getenv("PAGER"))
          strcpy(runcmd,getenv("PAGER"));
        else
          strcpy(runcmd,"less");
        break;
      
      case cmd_edit:
        if(getenv("EDITOR"))
          strcpy(runcmd,getenv("EDITOR"));
        else
          strcpy(runcmd,"vi");
        break;
      
      case cmd_other:
        /* prompt for command */
        mvaddstr(0,0,"Enter command to run, or none to abort:");
        echo(); mvgetstr(1,0,runcmd); noecho();
        if(runcmd[0]==0) cmdtype=cmd_none;
        break;
      
      default:
        mvaddstr(LINES-1,0,"can't happen");
      }
    
    if(cmdtype==cmd_none)
      {
      /* undo auto-selected file if need be */
      if(autosel) files[pos].tagged=0;
      showhdr();
      }
    else
      {
      char *ptr,sav;
      
      /* if we get here, they want to do it.
       * so, construct system() string to run.
       * first look for %s in cmdtext...
       */
      
      if((ptr=strstr(runcmd,"%s"))!=NULL)
        {
        int first=1;
        
        sav=*ptr; *ptr=0;
        strcpy(cmdtext,runcmd);
        for(f=0;f<numfiles;f++)
          if(files[f].tagged)
            {
            if(first) first=0; else strcat(cmdtext," ");
            strcat(cmdtext,files[f].name);
            }
        strcat(cmdtext,ptr+2);
        *ptr=sav;
        }
      else
        {
        /* ok, no %s. just cat the names on the end of the runcmd. */
        strcpy(cmdtext,runcmd);
        for(f=0;f<numfiles;f++)
          if(files[f].tagged)
            {
            strcat(cmdtext," ");
            strcat(cmdtext,files[f].name);
            }
        }
      
      /* add dest if it was defined */
      if(dest[0])
        strcat(cmdtext," "),strcat(cmdtext,dest);
      
      /* now quit, which will run the command */
      quit=1;
      }
    }
  }

*cmdtype_ptr=cmdtype;
}


void run_command(enum cmdtypes cmdtype,char *cmdtext)
{
int retval;

/* run any command */
switch(cmdtype)
  {
  case cmd_copy:
    printf("Copying files:\n"); break;
  case cmd_move:
    printf("Moving files:\n"); break;
  case cmd_delete:
    printf("Deleting files:\n"); break;
  case cmd_other: case cmd_view: case cmd_edit: case cmd_none:
    break;
  }

fflush(stdout);

if(cmdtype!=cmd_none)
  {
  retval=system(cmdtext);
  if(!opt_one && cmdtype!=cmd_edit && cmdtype!=cmd_view)
    {
    /* give status of command */
    if(retval)
      {
      char buf[128];
      printf("Command failed! Press return to continue: ");
      fflush(stdout);
      fgets(buf,sizeof(buf),stdin);
      }
    else
      {
      printf("Command finished OK.\n");
      fflush(stdout);
      sleep(1);
      }
    
    /* the actual repeating is done in main() */
    }
  }
}


void files_init()
{
int f;

/* clear tags */
for(f=0;f<MAX_FILES;f++)
  files[f].tagged=0;

printf("Reading/sorting dir...\n");

makefilelist(sorttype);

if(numfiles==0)
  {
  printf("No %sfiles found!\n",opt_all?"":"regular ");
  exit(1);
  }
}


int main(int argc,char *argv[])
{
enum cmdtypes cmdtype=cmd_none;
static char cmdtext[MAX_FILES*256];	/* used for constructing command */

parse_options(argc,argv);

/* this looks bizarre, but this inits curses then temp. 'escapes' from
 * it. We need to do this in case files_init() bombs out.
 */
curses_init(); curses_uninit();

do
  {
  /* read dir */
  files_init();
  curses_reinit();
  /* do the actual file selecting, and get command to run */
  do_select(&cmdtype,cmdtext);
  curses_uninit();
  run_command(cmdtype,cmdtext);
  }
while(!opt_one && cmdtype!=cmd_none);

return(0);
}
